// COB2MIDI.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <string.h>
#include "lzw.h"

int WriteVarLen (int value, unsigned char* outBuffer)
{
    int buffer;
	int index = 0;

    buffer = value & 0x7f;
    while ((value >>= 7) > 0)
    {
        buffer <<= 8;
        buffer |= 0x80;
        buffer += (value & 0x7f);
    }

    while (true)
    {
//        putc(buffer,outfile);
		outBuffer[index++] = buffer & 0xff;

        if (buffer & 0x80)
            buffer >>= 8;
        else
            break;
    }

	return index;
}

long ReadVarLen (FILE* infile)
{
    long value;
    unsigned char c;

    if ((value = getc(infile)) & 0x80)
    {
        value &= 0x7f;
        do
        {
                value = (value << 7) + ((c = getc(infile)) & 0x7f);
        } while (c & 0x80);
    }

    return value;
}

void writeHeader(FILE* outfile, unsigned short numberOfTracks)
{
	fwrite("MThd", 1, 4, outfile);

	// Header size
	fputc(0, outfile);
	fputc(0, outfile);
	fputc(0, outfile);
	fputc(6, outfile);

	// Format
	fputc(0, outfile);
	if (numberOfTracks <= 1)
		fputc(0, outfile);
	else
		fputc(1, outfile);

	// Number of tracks
	unsigned char* p = (unsigned char*)&numberOfTracks;
	fputc(p[1], outfile);
	fputc(p[0], outfile);

	// Division
	fputc(0, outfile);
	fputc(0x1b, outfile);
}

void writeTrack(FILE* outfile, int size)
{
	fwrite("MTrk", 1, 4, outfile);

	unsigned char* p = (unsigned char*)&size;
	fputc(p[3], outfile);
	fputc(p[2], outfile);
	fputc(p[1], outfile);
	fputc(p[0], outfile);

	fputc(0, outfile);
}

void writeTimeSignature(FILE* outfile)
{
	fputc(0xff, outfile);
	fputc(0x58, outfile);
	fputc(0x04, outfile);

	fputc(0x04, outfile);
	fputc(0x02, outfile);
	fputc(0x18, outfile);
	fputc(0x08, outfile);

	fputc(0x00, outfile);
}

void writeTempo(FILE* outfile)
{
	fputc(0xff, outfile);
	fputc(0x51, outfile);
	fputc(0x03, outfile);

	fputc(0x07, outfile);
	fputc(0x00, outfile);
	fputc(0x00, outfile);
}

#define TRACK_SIZE 100000
class TrackData
{
	public:
	int offset;
	unsigned char* data;

	TrackData()
	{
		data = new unsigned char[TRACK_SIZE];
		offset = 0;
	}
}; 

TrackData* channels[32];
int gnumberOfTracks = 0;

bool writeChannel(int channelNumber, int data, bool varLen)
{
	TrackData* track;

	if (channels[channelNumber] == 0)
	{

		track = channels[channelNumber] = new TrackData();
		gnumberOfTracks++;
	}
	else
	{
		track = channels[channelNumber];
		if (track->offset >= TRACK_SIZE)
			return false;
	}

	if (varLen && data > 0x7f)
		track->offset += WriteVarLen(data, &track->data[track->offset]); 
	else
		track->data[track->offset++] = data;

	return true;
}

struct NoteInfoStruct
{
	int offTime;
	unsigned char note;
	unsigned char velocity;
	unsigned char channelNumber;
	NoteInfoStruct* next;
};

class NoteQueue
{
public:
	NoteInfoStruct* head;
	int size;

	NoteQueue()
	{
		head = 0;
		size = 0;
	}

	~NoteQueue()
	{
		NoteInfoStruct* next = head;
		while (next != 0)
		{
			NoteInfoStruct* current = next;
			next = next->next;

			delete current;
		}
	}

	void put(int offTime, unsigned char note, unsigned char velocity, unsigned char channelNumber)
	{
		NoteInfoStruct* next = head;
		NoteInfoStruct* prev = next;
		while (next != 0)
		{
			prev = next;
			next = next->next;
		}

		next = new NoteInfoStruct;
		next->offTime = offTime;
		next->note = note;
		next->velocity = velocity;
		next->channelNumber = channelNumber;
		next->next = 0;

		if (head == 0)
			head = next;

		if (prev != 0)
			prev->next = next;

		size++;
	}

	NoteInfoStruct* get(int timeLine)
	{
		NoteInfoStruct* next = head;
		NoteInfoStruct* smallest = next;
		NoteInfoStruct* prevSmallest = next;
		
		// Find the note with the smallest note off time
		while (next != 0)
		{
			NoteInfoStruct* prev = next;
			next = next->next;
			if (next != 0)
			{
				if (smallest->offTime > next->offTime)
				{
					prevSmallest = prev;
					smallest = next;
				}
			}
		}

		if (smallest != 0)
		{
			if (smallest->offTime <= timeLine)
			{
				// Remove smallest
				if (smallest == head)
				{
					if (head->next != 0)
						head = head->next;
					else
						head = 0;
				}
				else if (prevSmallest->next != 0)
				{
					prevSmallest->next = prevSmallest->next->next;
				}

				size--;

				return smallest;
			}
		}

		return 0;
	}
};

void correctEndian(int* data, unsigned int size)
{
	unsigned char* p = (unsigned char*)data;

	for (unsigned int i = 0; i < size; i++)
	{
		data[i] = p[0] << 24 | p[1] << 16 | p[2] << 8 | p[3];
		p += 4;
	}
}

// Character mapping for Vietnamese from JBK to VNI-Times
int charLookup(unsigned char input, unsigned char* outBuffer)
{
	static unsigned char last = 0;

	int outSize = 0;
	switch (input)
	{
	case 0x5F:// A'
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xD9;
		break;

	case 0x60:// A`
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xD8;
		break;

	case 0x7B:// A(.
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xCB;
		break;

	case 0x7C:// A?
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xDB;
		break;

	case 0x7D:// A~
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xD5;
		break;

	case 0x7E:// A.
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xCF;
		break;

	case 0x80:// A^'
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xC1;
		break;

	case 0x81:// e?
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xFB;
		break;

	case 0x82:// e'
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xF9;
		break;

	case 0x83:// a^
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE2;
		break;

	case 0x84:// e~
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xF5;
		break;

	case 0x85:// a`
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xF8;
		break;

	case 0x86:// e.
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xEF;
		break;

	case 0x87:// A^`
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xC0;
		break;

	case 0x88:// e^
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xE2;
		break;

	case 0x89:// e^'
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xE1;
		break;

	case 0x8A:// e`
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xF8;
		break;

	case 0x8B:// e^`
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xE0;
		break;

	case 0x8D:// i`
		outBuffer[outSize++] = 0xEC;
		break;

	case 0x8E:// e^?
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xE5;
		break;

	case 0x8F:// e^~
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xE3;
		break;

	case 0x90:// A^~
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xCc;
		break;

	case 0x91:// o?
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xFB;
		break;

	case 0x92:// o~
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xF5;
		break;

	case 0x93:// o^
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xE2;
		break;

	case 0x94:// o.
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xEF;
		break;

	case 0x95:// o`
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xF8;
		break;

	case 0x96:// o^'
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xE1;
		break;

	case 0x97:// u`
		outBuffer[outSize++] = 'u';
		outBuffer[outSize++] = 0xF8;
		break;

	case 0x98:// o^`
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xE0;
		break;

	case 0x99:// o^?
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xE5;
		break;

	case 0x9A:// o^~
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xE3;
		break;

	case 0x9B:// o^.
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xE4;
		break;

	case 0x9C:// u?
		outBuffer[outSize++] = 'u';
		outBuffer[outSize++] = 0xFB;
		break;

	case 0x9D:// u~
		outBuffer[outSize++] = 'u';
		outBuffer[outSize++] = 0xF5;
		break;

	case 0x9E:// u.
		outBuffer[outSize++] = 'u';
		outBuffer[outSize++] = 0xEF;
		break;

	case 0x9F:// u*
		outBuffer[outSize++] = 0xF6;
		break;

	case 0xA0:// a'
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xF9;
		break;

	case 0xA1:// i'
		outBuffer[outSize++] = 'i';
		outBuffer[outSize++] = 0xF9;
		break;

	case 0xA2:// o'
		outBuffer[outSize++] = 'o';
		outBuffer[outSize++] = 0xF9;
		break;

	case 0xA3:// u'
		outBuffer[outSize++] = 'u';
		outBuffer[outSize++] = 0xF9;
		break;

	case 0xA4:// u*'
		outBuffer[outSize++] = 0xF6;
		outBuffer[outSize++] = 0xF9;
		break;

	case 0xA5:// u*`
		outBuffer[outSize++] = 0xF6;
		outBuffer[outSize++] = 0xF8;
		break;

	case 0xA6:// u*?
		outBuffer[outSize++] = 0xF6;
		outBuffer[outSize++] = 0xFB;
		break;

	case 0xA7:// u*~
		outBuffer[outSize++] = 0xF6;
		outBuffer[outSize++] = 0xF5;
		break;

	case 0xA8:// u*.
		outBuffer[outSize++] = 0xF6;
		outBuffer[outSize++] = 0xEF;
		break;

	case 0xA9:// i?
		outBuffer[outSize++] = 0xE6;
		break;

	case 0xAA:// i~
		outBuffer[outSize++] = 0xF3;
		break;

	case 0xAB:// i.
		outBuffer[outSize++] = 0xF2;
		break;

	case 0xAC:// e^.
		outBuffer[outSize++] = 'e';
		outBuffer[outSize++] = 0xE4;
		break;

	case 0xAD:// -d
		outBuffer[outSize++] = 0xF1;
		break;

	case 0xAE: // -D
		outBuffer[outSize++] = 0xD1;
		break;

	case 0xAF:// A^.
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xC4;
		break;

	case 0xB0:// A('
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xC9;
		break;

	case 0xB1:// A(`
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xC8;
		break;

	case 0xB2:// A(?
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xDA;
		break;

	case 0xB4:// E'
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xD9;
		break;

	case 0xB5:// E`
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xD8;
		break;

	case 0xB6:// E?
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xDB;
		break;

	case 0xB7:// E~
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xD5;
		break;

	case 0xB8:// E.
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xCF;
		break;

	case 0xB9:// E^'
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xC1;
		break;

	case 0xBA:// E^`
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xC0;
		break;

	case 0xBB:// E^?
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xC5;
		break;

	case 0xBC:// E^~
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xC3;
		break;

	case 0xBD:// E^.
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xC4;
		break;

	case 0xBE:// I'
		outBuffer[outSize++] = 0xCD;
		break;

	case 0xBF:// I`
		outBuffer[outSize++] = 0xCC;
		break;

	case 0xC0:// I?
		outBuffer[outSize++] = 0xC6;
		break;

	case 0xC1:// I~
		outBuffer[outSize++] = 0xD3;
		break;

	case 0xC2:// I.
		outBuffer[outSize++] = 0xD2;
		break;

	case 0xC3:// O'
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xD9;
		break;

	case 0xC4:// O`
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xD8;
		break;

	case 0xC5:// O?
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xDB;
		break;

	case 0xC6:// O~
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xD5;
		break;

	case 0xC7:// O.
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xCF;
		break;

	case 0xC8:// O^'
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xC1;
		break;

	case 0xC9:// O^`
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xC0;
		break;

	case 0xCA:// O^?
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xC5;
		break;

	case 0xCB:// O^~
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xC3;
		break;

	case 0xCC:// O^.
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xC4;
		break;

	case 0xCD:// O*'
		outBuffer[outSize++] = 0xD4;
		outBuffer[outSize++] = 0xD9;
		break;

	case 0xCE:// O*`
		outBuffer[outSize++] = 0xD4;
		outBuffer[outSize++] = 0xD8;
		break;

	case 0xCF:// O*?
		outBuffer[outSize++] = 0xD4;
		outBuffer[outSize++] = 0xDB;
		break;

	case 0xD0:// O*~
		outBuffer[outSize++] = 0xD4;
		outBuffer[outSize++] = 0xD5;
		break;

	case 0xD1:// O*.
		outBuffer[outSize++] = 0xD4;
		outBuffer[outSize++] = 0xCF;
		break;

	case 0xD2:// U'
		outBuffer[outSize++] = 'U';
		outBuffer[outSize++] = 0xD9;
		break;

	case 0xD3:// U`
		outBuffer[outSize++] = 'U';
		outBuffer[outSize++] = 0xD8;
		break;

	case 0xD4:// U?
		outBuffer[outSize++] = 'U';
		outBuffer[outSize++] = 0xDB;
		break;

	case 0xD5:// U~
		outBuffer[outSize++] = 'U';
		outBuffer[outSize++] = 0xD5;
		break;

	case 0xD6:// U.
		outBuffer[outSize++] = 'U';
		outBuffer[outSize++] = 0xCF;
		break;

	case 0xD7:// U*'
		outBuffer[outSize++] = 0xD6;
		outBuffer[outSize++] = 0xD9;
		break;

	case 0xD8:// U*`
		outBuffer[outSize++] = 0xD6;
		outBuffer[outSize++] = 0xD8;
		break;

	case 0xD9:// U*?
		outBuffer[outSize++] = 0xD6;
		outBuffer[outSize++] = 0xDB;
		break;

	case 0xDA:// U*~
		outBuffer[outSize++] = 0xD6;
		outBuffer[outSize++] = 0xD5;
		break;

	case 0xDB:// U*.
		outBuffer[outSize++] = 0xD6;
		outBuffer[outSize++] = 0xCF;
		break;

	case 0xDC:// Y'
		outBuffer[outSize++] = 'Y';
		outBuffer[outSize++] = 0xD9;
		break;

	case 0xDD:// Y`
		outBuffer[outSize++] = 'Y';
		outBuffer[outSize++] = 0xD8;
		break;

	case 0xDE:// Y?
		outBuffer[outSize++] = 'Y';
		outBuffer[outSize++] = 0xDB;
		break;

	case 0xDF:// Y~ 
		outBuffer[outSize++] = 'Y';
		outBuffer[outSize++] = 0xD5;
		break;

	case 0xE0:// a?
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xFB;
		break;

	case 0xE1:// a~
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] =0xF5;
		break;

	case 0xE2:// a.
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xEF;
		break;

	case 0xE3:// a^'
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE1;
		break;

	case 0xE4:// a^`
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE0;
		break;

	case 0xE5:// a^?
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE5;
		break;

	case 0xE6:// a^~
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE3;
		break;

	case 0xE7:// a^.
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE4;
		break;

	case 0xE8:// a(
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xEA;
		break;

	case 0xE9:// a('
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE9;
		break;

	case 0xEA:// a(`
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xE8;
		break;

	case 0xEB:// a(?
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xFA;
		break;

	case 0xEC:// a(~
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xFC;
		break;

	case 0xED:// a(.
		outBuffer[outSize++] = 'a';
		outBuffer[outSize++] = 0xEB;
		break;

	case 0xEE:// y'
		outBuffer[outSize++] = 'y';
		outBuffer[outSize++] = 0xF9;
		break;

	case 0xEF:// y`
		outBuffer[outSize++] = 'y';
		outBuffer[outSize++] = 0xF8;
		break;

	case 0xF0:// y?
		outBuffer[outSize++] = 'y';
		outBuffer[outSize++] = 0xFB;
		break;

	case 0xF1:// y~
		outBuffer[outSize++] = 'y';
		outBuffer[outSize++] = 0xF5;
		break;

	case 0xF2:// y.
		outBuffer[outSize++] = 0xFF;
		break;

	case 0xF3:// o*
		outBuffer[outSize++] = 0xF4;
		break;

	case 0xF4:// o*'
		outBuffer[outSize++] = 0xF4;
		outBuffer[outSize++] = 0xF9;
		break;

	case 0xF5: // o*`
		outBuffer[outSize++] = 0xF4;
		outBuffer[outSize++] = 0xF8;
		break;

	case 0xF6:// o*?
		outBuffer[outSize++] = 0xF4;
		outBuffer[outSize++] = 0xFB;
		break;

	case 0xF7:// o*~
		outBuffer[outSize++] = 0xF4;
		outBuffer[outSize++] = 0xF5;
		break;

	case 0xF8:// o*.
		outBuffer[outSize++] = 0xF4;
		outBuffer[outSize++] = 0xEF;
		break;

	case 0xF9:// O^
		outBuffer[outSize++] = 'O';
		outBuffer[outSize++] = 0xC2;
		break;

	case 0xFA:// O* 
		outBuffer[outSize++] = 0xD4;
		break;

	case 0xFB:// U*
		outBuffer[outSize++] = 0xD6;
		break;

	case 0xFC:// A(
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xCA;
		break;

	case 0xFD:// A^
		outBuffer[outSize++] = 'A';
		outBuffer[outSize++] = 0xC2;
		break;

	case 0xFE:// E^
		outBuffer[outSize++] = 'E';
		outBuffer[outSize++] = 0xC2;
		break;

	case 0xFF:// Y.
		outBuffer[outSize++] = 0xCE;
		break;

	default:
		outBuffer[outSize++] = input;
		if (input > 'z')
			printf("\n0x%02X%c 0x%02X\n", last, last, input);
	}

	last = input;

	return outSize;
}

//#define LOG_DATA

int cob2(int argc, char* argv[])
{
	FILE* infile;
	FILE* outfile;
#ifdef LOG_DATA
	FILE* log = fopen("log.txt", "w");
#endif
	int lyricCLength;
	int lyricULength;
	int midiCLength;
	int midiULength;
	unsigned char* lyricData;
	unsigned char* midiData;
	int midiOffset = 0;
	int lyricOffset = 0;
	unsigned char lyricChannelNumber = 0;
	unsigned char midiChannelNumber = 1;
	unsigned char* chunk;

	printf("COB (DTS Karaoke File Format) to MIDI Converter, version 1.0.0\n\n");

	if (argc < 2)
	{
		printf("Usage: COB2MIDI inputfilename outputfilename\n");
		return 0;
	}

	infile  = fopen(argv[1], "rb");
	if (infile == 0)
	{
		printf("Cannot open input file %s\n", argv[1]);
		return 0;
	}

	outfile = fopen(argv[2], "wb");
	if (outfile == 0)
	{
		printf("Cannot open output file %s\n", argv[2]);
		fclose(infile);
		return 0;
	}

	memset(channels, 0, sizeof(channels));

	// Get the lyric compressed length
	printf("Getting lyric compressed length... ");
	fseek(infile, 0, SEEK_SET);
	fread(&lyricCLength, 1, sizeof(lyricCLength), infile);
	correctEndian(&lyricCLength, 1);
	if (lyricCLength < 0)
	{
		printf("Bad lyric compressed length %08X\n", lyricCLength);
		fclose(infile);
		fclose(outfile);
		return 0;
	}
	printf("0x%08X\n", lyricCLength);

	// Get the lyric uncompressed length
	printf("Getting lyric uncompressed length... ");
	fseek(infile, 4, SEEK_SET);
	fread(&lyricULength, 1, sizeof(lyricULength), infile);
	correctEndian(&lyricULength, 1);
	if (lyricULength < 0)
	{
		printf("Bad lyric uncompressed length %08X\n", lyricULength);
		fclose(infile);
		fclose(outfile);
		return 0;
	}
	printf("0x%08X\n", lyricULength);

	// Read the lyric part into a buffer
	printf("Reading compressed lyric part into a buffer... ");
	unsigned char* tempBuffer = new unsigned char[lyricCLength];
	fseek(infile, 12, SEEK_SET);
	fread(tempBuffer, 1, lyricCLength, infile);
	printf("Done\n");
/*
	// Test code
	FILE* fpTest = fopen("00001_lyric2.com", "wb");
	fwrite(tempBuffer, 1, lyricCLength, fpTest);
	fclose(fpTest);
*/
	// Decompress the lyric part
	printf("Decompressing the lyric part... ");
	lyricData = new unsigned char[lyricULength + 10];
	int tempLength = expand(tempBuffer, lyricCLength, lyricData);
	if (tempLength < lyricULength)
	{
		printf("Uncompressed lyric sizes don't match expected: 0x%08X actual: 0x%08x\n", lyricULength, tempLength);
	}
	delete[] tempBuffer;
	printf("Done\n");

	// Converting the lyric format into MIDI
	printf("Converting the lyric into MIDI format");
	int lastDelay = 0;
	bool firstTime = true;
	unsigned char translatedBuffer[5];
	int testcount = 0;
	do
	{
		chunk = &lyricData[lyricOffset];
		lyricOffset += 4;

		// Test code
		testcount++;
		if (testcount <= 41)
			continue;

		int delay = chunk[3] << 8 | chunk[2];
		if (delay > 0)
		{
			writeChannel(lyricChannelNumber, delay - lastDelay, true);
			lastDelay = delay;
		}
		else
		{
			writeChannel(lyricChannelNumber, 0, true);
		}

		writeChannel(lyricChannelNumber, 0xff, false);
		writeChannel(lyricChannelNumber, 0x05, false);
		
		int transSize = charLookup(chunk[0], translatedBuffer);
		writeChannel(lyricChannelNumber, transSize, false);
		for (int i = 0; i < transSize; i++)
			writeChannel(lyricChannelNumber, translatedBuffer[i], false);
		
		if ((lyricOffset % 40) == 0)
			printf(".");
	}
	while (lyricOffset < lyricULength);
	printf("Done\n");

	// Get the MIDI compressed length
	printf("Getting MIDI compressed length... ");
	fseek(infile, lyricCLength + 12, SEEK_SET);
	fread(&midiCLength, 1, sizeof(midiCLength), infile);
	correctEndian(&midiCLength, 1);
	if (midiCLength < 0)
	{
		printf("Bad MIDI compressed length %08X\n", midiCLength);
		fclose(infile);
		fclose(outfile);
		return 0;
	}
	printf("0x%08X\n", midiCLength);

	// Get the MIDI uncompressed length
	printf("Getting MIDI uncompressed length... ");
	fseek(infile, lyricCLength + 16, SEEK_SET);
	fread(&midiULength, 1, sizeof(midiULength), infile);
	correctEndian(&midiULength, 1);
	if (midiULength < 0)
	{
		printf("Bad MIDI uncompressed length %08X\n", midiULength);
		fclose(infile);
		fclose(outfile);
		return 0;
	}
	printf("0x%08X\n", midiULength);

	// Read the MIDI part into a buffer
	printf("Reading the MIDI compress data into a buffer... ");
	tempBuffer = new unsigned char[midiCLength];
	fseek(infile, lyricCLength + 12 + 12, SEEK_SET);
	fread(tempBuffer, 1, midiCLength, infile);
	printf("Done\n");
/*
	// Test code
	FILE* fpTest = fopen("00001_midi2.com", "wb");
	fwrite(tempBuffer, 1, midiCLength, fpTest);
	fclose(fpTest);
*/
	// Decompress the MIDI part
	printf("Decompressing the MIDI part... ");
	midiData = new unsigned char[midiULength + 10];
	tempLength = expand(tempBuffer, midiCLength, midiData);
	if (tempLength < midiULength)
	{
		printf("Uncompressed MIDI sizes don't match expected: 0x%08X actual: 0x%08x\n", midiULength, tempLength);
	}
	delete[] tempBuffer;
	printf("Done\n");
	
	// Test code
	FILE* fpDebug = fopen("00001_midi2.unc", "wb");
	fwrite(midiData, 1, midiULength, fpDebug);
	fclose(fpDebug);


	// Decode and write the MIDI part to a buffer
	printf("Converting the MIDI part to MIDI format");
	NoteQueue notes;
	int timeLine = 0;
	do
	{
		chunk = &midiData[midiOffset];
		midiOffset += 5;

#ifdef LOG_DATA
//		if (midiChannelNumber == 0)
			fprintf(log, "%d %d 0x%02X 0x%02X 0x%02X\n", chunk[0], chunk[1], chunk[2], chunk[3], chunk[4]);
#endif

		if (chunk[0] > 0)
		{
			int nextTimeLine = timeLine + chunk[0];

			// Process notes off
			NoteInfoStruct* note;
			int tempTileLine = timeLine;
			int tempChunk0 = chunk[0];
			while ((note = notes.get(nextTimeLine)) != 0)
			{
				int timeDif = note->offTime - tempTileLine;
				if (timeDif < 0 || timeDif > 300)
					printf("bad\n");

				tempTileLine += timeDif;
				tempChunk0 -= timeDif;
				writeChannel(midiChannelNumber, timeDif, true); // rest/delay

				writeChannel(midiChannelNumber,  0x80 | note->channelNumber, false);// note on/off
				writeChannel(midiChannelNumber, note->note, false); // pitch/note
				writeChannel(midiChannelNumber, note->velocity, false); // velocity

				delete note;
			}

			writeChannel(midiChannelNumber, tempChunk0, true); // rest/delay
			timeLine += chunk[0];
		}
		else
		{
			writeChannel(midiChannelNumber, 0, true); // rest/delay
		}

		switch (chunk[2] & 0xf0)
		{
		case 0xB0:
		case 0xE0:
			writeChannel(midiChannelNumber, chunk[2], false);
			writeChannel(midiChannelNumber, chunk[3], false);
			writeChannel(midiChannelNumber, chunk[4], false);
			break;

		case 0xC0:
		case 0xD0:
			writeChannel(midiChannelNumber, chunk[2], false);
			writeChannel(midiChannelNumber, chunk[3], false);
			break;

		case 0x00:
			// Note on
			writeChannel(midiChannelNumber,  0x90 | (chunk[2] & 0xf), false);// note on/off
			writeChannel(midiChannelNumber, chunk[3], false); // pitch/note
			writeChannel(midiChannelNumber, chunk[4], false); // velocity

			// Remember the note for note off processing
			notes.put(timeLine + chunk[1], chunk[3], chunk[4], chunk[2] & 0xf);
			break;

		case 0xF0:
			// We'll write the data as is to the buffer
			midiOffset -= 3;
			while (midiData[midiOffset] != 0xF7)
			{
//				writeChannel(midiChannelNumber, midiData[midiOffset], false);
				midiOffset++;
			}
//			writeChannel(midiChannelNumber, midiData[midiOffset], false);
			midiOffset++;
			printf("Unrecognize code %02X\n", chunk[2]);
			break;

		default:
			printf("Unrecognize code %02X\n", chunk[2]);
		}

		if ((midiOffset % 50) == 0)
			printf(".");
	}
	while (midiOffset < midiULength);
	printf("Done\n");

	if (notes.size > 0)
	{
		// Process notes off
		NoteInfoStruct* note;
		int tempTileLine = timeLine;
		while ((note = notes.get(timeLine + 1000)) != 0)
		{
			int timeDif = note->offTime - tempTileLine;
			if (timeDif < 0 || timeDif > 300)
				printf("bad\n");

			tempTileLine += timeDif;
			writeChannel(midiChannelNumber, timeDif, true); // rest/delay

			writeChannel(midiChannelNumber,  0x80 | note->channelNumber, false);// note on/off
			writeChannel(midiChannelNumber, note->note, false); // pitch/note
			writeChannel(midiChannelNumber, note->velocity, false); // velocity

			delete note;
		}
	}

	printf("Writing the MIDI buffer to the file... ");
	if (gnumberOfTracks > 0)
	{
		writeHeader(outfile, gnumberOfTracks);

		// Write the tracks
		for (int i = 0; i < 32; i++)
		{
			TrackData* track = channels[i];
			if (track != 0)
			{
				writeTrack(outfile, track->offset + 19);

				writeTimeSignature(outfile);
				writeTempo(outfile);

				fwrite(track->data, 1, track->offset, outfile);

				fputc(0x00, outfile);
				fputc(0xff, outfile);
				fputc(0x2f, outfile);
				fputc(0x00, outfile);

				delete[] track->data;
				delete track;
			}
		}
	}
	printf("Done\n");

	delete[] lyricData;
	delete[] midiData;

	fclose(infile);
	fclose(outfile);

#ifdef LOG_DATA
	fclose(log);
#endif

	return 0;
}
