// MIDI2COB.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include "lzw.h"

#pragma pack(1)
struct MTHD_CHUNK
{
   /* Here's the 8 byte header that all chunks must have */
   char           ID[4];  /* This will be 'M','T','h','d' */
   unsigned int   Length; /* This will be 6 */

   /* Here are the 6 bytes */
   unsigned short Format;
   unsigned short NumTracks;
   unsigned short Division;
};

struct MTRK_CHUNK
{
   /* Here's the 8 byte header that all chunks must have */
   char           ID[4];   /* This will be 'M','T','r','k' */
   unsigned int   Length;  /* This will be the actual size of Data[] */

   /* Here are the data bytes */
   unsigned char  Data[1];  /* Its actual size is Data[Length] */
};

void correctEndian(unsigned int* data, unsigned int size)
{
	unsigned char* p = (unsigned char*)data;

	for (unsigned int i = 0; i < size; i++)
	{
		data[i] = p[0] << 24 | p[1] << 16 | p[2] << 8 | p[3];
		p += 4;
	}
}

void correctEndian(unsigned short* data, unsigned int size)
{
	unsigned char* p = (unsigned char*)data;

	for (unsigned int i = 0; i < size; i++)
	{
		data[i] = p[0] << 8 | p[1];
		p += 2;
	}
}

long ReadVarLen (unsigned char* inBuffer, unsigned int* offset)
{
    long value;
    unsigned char c;

    if ((value = inBuffer[(*offset)++]) & 0x80)
    {
        value &= 0x7f;
        do
        {
                value = (value << 7) + ((c = inBuffer[(*offset)++]) & 0x7f);
        } while (c & 0x80);
    }

    return value;
}

struct EventInfoStruct
{
	int time;
	unsigned char* data;
	EventInfoStruct* next;
};

class ChannelEventQueue
{
public:
	EventInfoStruct* head;
	int size;

	ChannelEventQueue()
	{
		head = 0;
		size = 0;
	}

	~ChannelEventQueue()
	{
		EventInfoStruct* next = head;
		while (next != 0)
		{
			EventInfoStruct* current = next;
			next = next->next;

			delete current;
		}
	}

	void put(int time, unsigned char* data)
	{
		EventInfoStruct* next = head;
		EventInfoStruct* prev = next;
		while (next != 0)
		{
			prev = next;
			next = next->next;
		}

		next = new EventInfoStruct;
		next->time = time;
		next->data = data;
		next->next = 0;

		if (head == 0)
			head = next;

		if (prev != 0)
			prev->next = next;

		size++;
	}

	EventInfoStruct* get()
	{
		EventInfoStruct* next = head;
		EventInfoStruct* smallest = next;
		EventInfoStruct* prevSmallest = next;
		
		// Find the note with the smallest note off time
		while (next != 0)
		{
			EventInfoStruct* prev = next;
			next = next->next;
			if (next != 0)
			{
				if (smallest->time > next->time)
				{
					prevSmallest = prev;
					smallest = next;
				}
			}
		}

		if (smallest != 0)
		{
			// Remove smallest
			if (smallest == head)
			{
				if (head->next != 0)
					head = head->next;
				else
					head = 0;
			}
			else if (prevSmallest->next != 0)
			{
				prevSmallest->next = prevSmallest->next->next;
			}

			size--;

			return smallest;
		}

		return 0;
	}
};

void dumpText(unsigned char* buffer, int length)
{
	// Dump text
	for (int i = 0; i < length; i++)
	{
		printf("%c", buffer[i]);
	}
	printf("\n");
}

struct CobChunkStruct
{
	int eventTime;
	int noteDuration;
	unsigned char command;
	unsigned char note;
	unsigned char velocity;
	CobChunkStruct* next;
};

class CobQueue
{
public:
	CobChunkStruct* head;
	int size;

	CobQueue()
	{
		head = 0;
		size = 0;
	}

	~CobQueue()
	{
		CobChunkStruct* next = head;
		while (next != 0)
		{
			CobChunkStruct* current = next;
			next = next->next;

			delete current;
		}
	}

	void put(int eventTime, unsigned char command, unsigned char note, unsigned char velocity)
	{
		CobChunkStruct* next = head;
		CobChunkStruct* prev = next;
		while (next != 0)
		{
			prev = next;
			next = next->next;
		}

		next               = new CobChunkStruct;
		next->eventTime    = eventTime;
		next->noteDuration = 0;
		next->command      = command;
		next->note         = note;
		next->velocity     = velocity;
		next->next         = 0;

		if (head == 0)
			head = next;

		if (prev != 0)
			prev->next = next;

		size++;
	}

	CobChunkStruct* find(unsigned char command, unsigned char note)
	{
		CobChunkStruct* next = head;
		
		while (next != 0)
		{
			if (next->noteDuration == 0 &&
				next->command == command && 
				next->note == note)
			{
				return next;
			}

			next = next->next;
		}

		return next;
	}

	CobChunkStruct* get()
	{
		CobChunkStruct* firstIn = head;

		if (head != 0)
		{
			head = head->next;
			size--;
		}

		return firstIn;
	}
};


struct CobLyricChunkStruct
{
	int eventTime;
	unsigned char syl;
	CobLyricChunkStruct* next;
};

class CobLyricQueue
{
public:
	CobLyricChunkStruct* head;
	int size;

	CobLyricQueue()
	{
		head = 0;
		size = 0;
	}

	~CobLyricQueue()
	{
		CobLyricChunkStruct* next = head;
		while (next != 0)
		{
			CobLyricChunkStruct* current = next;
			next = next->next;

			delete current;
		}
	}

	void put(int eventTime, unsigned char syl)
	{
		CobLyricChunkStruct* next = head;
		CobLyricChunkStruct* prev = next;
		while (next != 0)
		{
			prev = next;
			next = next->next;
		}

		next            = new CobLyricChunkStruct;
		next->eventTime = eventTime;
		next->syl       = syl;
		next->next      = 0;

		if (head == 0)
			head = next;

		if (prev != 0)
			prev->next = next;

		size++;
	}

	CobLyricChunkStruct* get()
	{
		CobLyricChunkStruct* firstIn = head;

		if (head != 0)
		{
			head = head->next;
			size--;
		}

		return firstIn;
	}
};

unsigned char** asciiJBKLookup;

int convertASCIIToJBKFont(unsigned char input, bool firstTime)
{
	static unsigned char lastChar;
	int output = 0;

	if (firstTime)
	{
		lastChar = 0;
	}

	if (input == 0)
	{
		return lastChar;
	}

	if (lastChar != 0)
	{
		if (asciiJBKLookup[lastChar][input] != 0)
		{
			lastChar = asciiJBKLookup[lastChar][input];
		}
		else
		{
			output = lastChar;
			lastChar = input;
		}
	}
	else
	{
		lastChar = input;
	}

	return output;
}

void outTest(unsigned char* output, int* offset, unsigned char c, bool firstTime)
{
	static unsigned char last4Bits = 0;
	static bool bitRemain = false;

	if (firstTime)
	{
		last4Bits = 0;
		bitRemain = 0;
	}

	if (bitRemain == false)
	{
		int out = (c >> 4);
		output[(*offset)++] = out;

		last4Bits = c & 0xf;
		bitRemain = true;
	}
	else
	{
		int out = (last4Bits << 4);
		output[(*offset)++] = out;
		output[(*offset)++] = c;
		bitRemain = false;
	}
}

int compressTest(unsigned char* input, int inputSize, unsigned char* output)
{
	int compressSize = 0;
	bool firstTime = true;

	for (int i = 0; i < inputSize; i++)
	{
		outTest(output, &compressSize, input[i], firstTime);
		firstTime = false;

		if (i == 0xfff)
			printf("****");

		if ((i % 4) == 0 && i > 0xf00 && i < 0x10ff)
			printf("%c ", input[i]);
	}

	return compressSize;
}

int midi2Cob(char* midiFileName, FILE* outFile)
{
	int fileSize = 0;
	FILE* fp = fopen(midiFileName, "rb");
	MTHD_CHUNK header;
	unsigned char** tracks;
	unsigned int* trackOffsets;

	if (fp == 0)
		return 0;

    // Initialize the lookup table
	asciiJBKLookup = new unsigned char*[256];
	for (int i = 0; i < 256; i++)
	{
		asciiJBKLookup[i] = new unsigned char[256];
		memset ( asciiJBKLookup[i], 0, 256);
	}

	// ************************** Lookup table ASCII-JBK *********************
    // A
    asciiJBKLookup['A']['\''] = 0x5F;
    asciiJBKLookup['A']['`'] = 0x60;
    asciiJBKLookup['A']['?'] = 0x7C; 
    asciiJBKLookup['A']['~'] = 0x7D;
    asciiJBKLookup['A']['.'] = 0x7E;
    asciiJBKLookup['A']['('] = 0xFC; 
    asciiJBKLookup['A']['^'] = 0xFD;
    
	// a
    asciiJBKLookup['a']['\''] = 0xA0;
    asciiJBKLookup['a']['`'] = 0x85;
    asciiJBKLookup['a']['?'] = 0xE0; 
    asciiJBKLookup['a']['~'] = 0xE1;
    asciiJBKLookup['a']['.'] = 0xE2;
    asciiJBKLookup['a']['('] = 0xE8;
    asciiJBKLookup['a']['^'] = 0x83;

	// A(
	asciiJBKLookup[0xFC]['\''] = 0xB0;
    asciiJBKLookup[0xFC]['`'] = 0xB1;
    asciiJBKLookup[0xFC]['?'] = 0xB2;
    asciiJBKLookup[0xFC]['.'] = 0x7B;  

	// a(
	asciiJBKLookup[0xE8]['\''] = 0xE9;
    asciiJBKLookup[0xE8]['`'] = 0xEA;
    asciiJBKLookup[0xE8]['?'] = 0xEB;
    asciiJBKLookup[0xE8]['~'] = 0xEC;
    asciiJBKLookup[0xE8]['.'] = 0xED;  

	// A^
	asciiJBKLookup[0xFD]['\''] = 0x80;
    asciiJBKLookup[0xFD]['`'] = 0x87;
//    asciiJBKLookup[0xFD]['?'] = 0xEB; don't know
    asciiJBKLookup[0xFD]['~'] = 0x90;
    asciiJBKLookup[0xFD]['.'] = 0xAF;  

	// a^
	asciiJBKLookup[0x83]['\''] = 0xE3;
    asciiJBKLookup[0x83]['`'] = 0xE4; 
    asciiJBKLookup[0x83]['?'] = 0xE5;
    asciiJBKLookup[0x83]['~'] = 0xE6;
    asciiJBKLookup[0x83]['.'] = 0xE7; 

	// D
    asciiJBKLookup['D']['D'] = 0xAE; 	
    asciiJBKLookup['D']['d'] = 0xAE; 	

	// d
    asciiJBKLookup['d']['d'] = 0xAD; 	

	// E
    asciiJBKLookup['E']['\''] = 0xB4;
    asciiJBKLookup['E']['`'] = 0xB5;
    asciiJBKLookup['E']['?'] = 0xB6; 
    asciiJBKLookup['E']['~'] = 0xB7;
    asciiJBKLookup['E']['.'] = 0xB8;
    asciiJBKLookup['E']['^'] = 0xFE;
    
    // e
	asciiJBKLookup['e']['\''] = 0x82;
    asciiJBKLookup['e']['`'] = 0x8A;
    asciiJBKLookup['e']['?'] = 0x81; 
    asciiJBKLookup['e']['~'] = 0x84;
    asciiJBKLookup['e']['.'] = 0x86;
    asciiJBKLookup['e']['^'] = 0x88;

	// E^
    asciiJBKLookup[0xFE]['\''] = 0xB9;
    asciiJBKLookup[0xFE]['`'] = 0xBA;
    asciiJBKLookup[0xFE]['?'] = 0xBB; 
    asciiJBKLookup[0xFE]['~'] = 0xBC;
    asciiJBKLookup[0xFE]['.'] = 0xBD;

	// e^
	asciiJBKLookup[0x88]['\''] = 0x89;
    asciiJBKLookup[0x88]['`'] = 0x8B;
    asciiJBKLookup[0x88]['?'] = 0x8E;
    asciiJBKLookup[0x88]['~'] = 0x8F;
    asciiJBKLookup[0x88]['.'] = 0xAC;

	// I
	asciiJBKLookup['I']['\''] = 0xBE;
    asciiJBKLookup['I']['`'] = 0xBF;
    asciiJBKLookup['I']['?'] = 0xC0;
    asciiJBKLookup['I']['~'] = 0xC1;
    asciiJBKLookup['I']['.'] = 0xC2;

	// i
	asciiJBKLookup['i']['\''] = 0xA1;
    asciiJBKLookup['i']['`'] = 0x8D;
    asciiJBKLookup['i']['?'] = 0xA9;
    asciiJBKLookup['i']['~'] = 0xAA;
    asciiJBKLookup['i']['.'] = 0xAB;

    // O
    asciiJBKLookup['O']['\''] = 0xC3;
    asciiJBKLookup['O']['`'] = 0xC4;
    asciiJBKLookup['O']['?'] = 0xC5; 
    asciiJBKLookup['O']['~'] = 0xC6;
    asciiJBKLookup['O']['.'] = 0xC7;
    asciiJBKLookup['O']['^'] = 0xF9;
    asciiJBKLookup['O']['*'] = 0xFC;

    // o
    asciiJBKLookup['o']['\''] = 0xa2;
    asciiJBKLookup['o']['`'] = 0x95;
    asciiJBKLookup['o']['?'] = 0x91; 
    asciiJBKLookup['o']['~'] = 0x92;
    asciiJBKLookup['o']['.'] = 0x94;
    asciiJBKLookup['o']['^'] = 0x93;
    asciiJBKLookup['o']['*'] = 0xF3;

    // O^
    asciiJBKLookup[0xF9]['\''] = 0xC8;
    asciiJBKLookup[0xF9]['`'] = 0xC9;
    asciiJBKLookup[0xF9]['?'] = 0xCA; 
    asciiJBKLookup[0xF9]['~'] = 0xCB;
    asciiJBKLookup[0xF9]['.'] = 0xCC;

	// o^
    asciiJBKLookup[0x93]['\''] = 0x96;
    asciiJBKLookup[0x93]['`'] = 0x98;
    asciiJBKLookup[0x93]['?'] = 0x99; 
    asciiJBKLookup[0x93]['~'] = 0x9A;
    asciiJBKLookup[0x93]['.'] = 0x9B;

	// O*
    asciiJBKLookup[0xFC]['\''] = 0xCD;
    asciiJBKLookup[0xFC]['`'] = 0xCE;
    asciiJBKLookup[0xFC]['?'] = 0xCF; 
    asciiJBKLookup[0xFC]['~'] = 0xD0;
    asciiJBKLookup[0xFC]['.'] = 0xD1;

	// o*
    asciiJBKLookup[0xF3]['\''] = 0xF4;
    asciiJBKLookup[0xF3]['`'] = 0xF5;
    asciiJBKLookup[0xF3]['?'] = 0xF6; 
    asciiJBKLookup[0xF3]['~'] = 0xF7;
    asciiJBKLookup[0xF3]['.'] = 0xF8;

    // U
    asciiJBKLookup['U']['\''] = 0xD2;
    asciiJBKLookup['U']['`'] = 0xD3;
    asciiJBKLookup['U']['?'] = 0xD4; 
    asciiJBKLookup['U']['~'] = 0xD5;
    asciiJBKLookup['U']['.'] = 0xD6;
    asciiJBKLookup['U']['*'] = 0xFB;

    // u
    asciiJBKLookup['u']['\''] = 0xA3;
    asciiJBKLookup['u']['`'] = 0x97;
    asciiJBKLookup['u']['?'] = 0x9c; 
    asciiJBKLookup['u']['~'] = 0x9d;
    asciiJBKLookup['u']['.'] = 0x9e;
    asciiJBKLookup['u']['*'] = 0x9F;

    // U*
    asciiJBKLookup[0xFB]['\''] = 0xD7;
    asciiJBKLookup[0xFB]['`'] = 0xD8;
    asciiJBKLookup[0xFB]['?'] = 0xD9; 
    asciiJBKLookup[0xFB]['~'] = 0xDA;
    asciiJBKLookup[0xFB]['.'] = 0xDB;

    // u*
    asciiJBKLookup[0x9F]['\''] = 0xA4;
    asciiJBKLookup[0x9F]['`'] = 0xA5;
    asciiJBKLookup[0x9F]['?'] = 0xA6; 
    asciiJBKLookup[0x9F]['~'] = 0xA7;
    asciiJBKLookup[0x9F]['.'] = 0xA8;

    // Y
    asciiJBKLookup['Y']['\''] = 0xDC;
    asciiJBKLookup['Y']['`'] = 0xDD;
    asciiJBKLookup['Y']['?'] = 0xDE; 
    asciiJBKLookup['Y']['.'] = 0xFF; 

	// y
    asciiJBKLookup['y']['\''] = 0xEE;
    asciiJBKLookup['y']['`'] = 0xEF;
    asciiJBKLookup['y']['?'] = 0xF0; 
    asciiJBKLookup['y']['~'] = 0xF1;
    asciiJBKLookup['y']['.'] = 0xF2;
	// ************************** Lookup table *********************

	// Read header
	fread(&header, 1, sizeof(header), fp);
	correctEndian(&header.Length, 1);
	correctEndian(&header.Format, 1);
	correctEndian(&header.Division, 1);
	correctEndian(&header.NumTracks, 1);

	tracks = new unsigned char*[header.NumTracks];
	trackOffsets = new unsigned int[header.NumTracks];
	memset(trackOffsets, 0, header.NumTracks * sizeof(int));

	for (i = 0; i < header.NumTracks; i++)
	{
		// Read the track header
		MTRK_CHUNK trackHeader;
		fread(&trackHeader, 1, offsetof(MTRK_CHUNK, Data), fp);
		correctEndian(&trackHeader.Length, 1);

		// Read the track data
		tracks[i] = new unsigned char[offsetof(MTRK_CHUNK, Data) + trackHeader.Length];
		memcpy(tracks[i], &trackHeader, offsetof(MTRK_CHUNK, Data));
		fread(&tracks[i][offsetof(MTRK_CHUNK, Data)], 1, trackHeader.Length, fp);
	}	

	fclose(fp);

	// Extract tracks events and put into a queue
	ChannelEventQueue channelEventQueue;
	int maxTimeLine = 0;
	for (i = 0; i < header.NumTracks; i++)
	{
		int channelTimeLine = 0;
		MTRK_CHUNK* trackHeader = (MTRK_CHUNK*)tracks[i];

		while (trackOffsets[i] < trackHeader->Length)
		{
			// Get channel delay time
			int time = ReadVarLen(trackHeader->Data, &trackOffsets[i]);

			// Put the event into a queue
			channelEventQueue.put(channelTimeLine + time, &trackHeader->Data[trackOffsets[i]]);

			channelTimeLine += time;

			// Find out the event length
			unsigned char command = trackHeader->Data[trackOffsets[i]++];
			switch (command)
			{
			case 0xFF:
				command = trackHeader->Data[trackOffsets[i]++];
				switch (command)
				{
				case 0x00:// FF 00 02 ssss Sequence Number 
					trackOffsets[i] += 2;
					break;

				case 0x01:// FF 01 len text Text Event 
				case 0x02:// FF 02 len text Copyright Notice 
				case 0x03:// FF 03 len text Sequence/Track Name 
				case 0x04:// FF 04 len text Instrument Name 
				case 0x05:// FF 05 len text Lyric 
				case 0x06:// FF 06 len text Marker 
				case 0x07:// FF 07 len text Cue Point 
					trackOffsets[i] += ReadVarLen(trackHeader->Data, &trackOffsets[i]);
					break;

				case 0x20:// FF 20 01 0n Channel Prefix
					trackOffsets[i] += 2;
					break;

				case 0x21:// FF 21 01 0n Port Prefix 
					trackOffsets[i] += 2;
					break;

				case 0x2F:// FF 2F 00 End of Track 
					trackOffsets[i] += 1;
					break;

				case 0x51:// FF 51 03 tttttt Set Tempo, in microseconds per MIDI quarter-note 
					trackOffsets[i] += 4;
					break;

				case 0x54:// FF 54 05 hr mn se fr ff SMPTE Offset (New in 0.06 - SMPTE Format specification)
					trackOffsets[i] += 6;
					break;

				case 0x58:// FF 58 04 nn dd cc bb Time Signature 
					trackOffsets[i] += 5;
					break;

				case 0x59:// FF 59 02 sf mi Key Signature 
					trackOffsets[i] += 3;
					break;

				case 0x7F:// FF 7F len data Sequencer-Specific Meta-Event 
					trackOffsets[i] +=  ReadVarLen(trackHeader->Data, &trackOffsets[i]);
					break;

				default:
					printf("Unknown command 0xFF 0x%02X\n", command); 
				}
				break;

			case 0xF0:
			case 0xF7:
				printf("Unknown command 0x%02X, %04X\n", command, trackOffsets[i]); 
				break;

			default:
				switch (command & 0xF0)
				{
				case 0x80:
					trackOffsets[i] += 2;
					break;

				case 0x90:
					trackOffsets[i] += 2;
					break;

				case 0xA0:
					trackOffsets[i] += 2;
					break;

				case 0xB0:
					trackOffsets[i] += 2;
					break;

				case 0xC0:
					trackOffsets[i] += 1;
					break;

				case 0xD0:
					trackOffsets[i] += 1;
					break;

				case 0xE0:
					trackOffsets[i] += 2;
					break;

				default:
					printf("Unknown command 0x%02X, %04X\n", command, trackOffsets[i]); 
				}
			}
		}

		if (channelTimeLine > maxTimeLine)
			maxTimeLine = channelTimeLine;
	}

	// Process the events
	EventInfoStruct* eventInfo;
	CobQueue cobQueue;
	CobLyricQueue lyricQueue;
	unsigned int temp;
	int tempo = 500000;// default to 120 BPM
	EventInfoStruct* prevLyricEventInfo = 0;
	while ((eventInfo = channelEventQueue.get()) != 0)
	{
		unsigned char command = eventInfo->data[0];
		int textLength;
		switch (command)
		{
		case 0xFF:
			command = eventInfo->data[1];
			switch (command)
			{
			case 0x00:// FF 00 02 ssss Sequence Number 
				break;

			case 0x01:// FF 01 len text Text Event 
			case 0x02:// FF 02 len text Copyright Notice 
			case 0x03:// FF 03 len text Sequence/Track Name 
			case 0x04:// FF 04 len text Instrument Name 
				break;

			case 0x05:// FF 05 len text Lyric 
				{
					if (prevLyricEventInfo != 0)
					{
						temp = 2;
						textLength = ReadVarLen(prevLyricEventInfo->data, &temp);

						float timePerSyl = (float)(eventInfo->time - prevLyricEventInfo->time) / textLength;
						for (int i = 0; i < textLength; i++)
						{
							lyricQueue.put((int)(prevLyricEventInfo->time + i * timePerSyl), 
								prevLyricEventInfo->data[temp + i]);
						}

//						dumpText(&prevLyricEventInfo->data[temp], textLength);
					}

					prevLyricEventInfo = eventInfo;
/*
					temp = 2;
					textLength = ReadVarLen(eventInfo->data, &temp);

					for (int i = 0; i < textLength; i++)
					{
						lyricQueue.put(eventInfo->time, eventInfo->data[temp + i]);
					}
*/
				}
				break;

			case 0x06:// FF 06 len text Marker 
			case 0x07:// FF 07 len text Cue Point 
				break;

			case 0x20:// FF 20 01 0n Channel Prefix
				break;

			case 0x21:// FF 21 01 0n Port Prefix 
				break;

			case 0x2F:// FF 2F 00 End of Track 
				break;

			case 0x51:// FF 51 03 tttttt Set Tempo, in microseconds per MIDI quarter-note 
				tempo = eventInfo->data[3] << 16 | eventInfo->data[4] << 8 | eventInfo->data[5];
				break;

			case 0x54:// FF 54 05 hr mn se fr ff SMPTE Offset (New in 0.06 - SMPTE Format specification)
				break;

			case 0x58:// FF 58 04 nn dd cc bb Time Signature 
				printf("Time Signature %d %d %d %d\n", eventInfo->data[3], eventInfo->data[4], eventInfo->data[5], eventInfo->data[5]);
				break;

			case 0x59:// FF 59 02 sf mi Key Signature 
				printf("Key Signature %d %d\n", eventInfo->data[3], eventInfo->data[4]);
				break;

			case 0x7F:// FF 7F len data Sequencer-Specific Meta-Event 
				break;

			default:
				printf("Unknown command 0xFF 0x%02X\n", command); 
			}
			break;

		case 0xF0:
		case 0xF7:
			printf("Unknown command 0x%02X, %04X\n", command, trackOffsets[i]); 
			break;

		default:
			switch (command & 0xF0)
			{
			case 0x80:
				{
					CobChunkStruct* cobInfo;
					while ((cobInfo = cobQueue.find(eventInfo->data[0]  & 0x0F, eventInfo->data[1])) != 0)
					{
						cobInfo->noteDuration = eventInfo->time - cobInfo->eventTime;
					}
				}
				break;

			case 0x90:
				if (eventInfo->data[2] > 0) // if velocity is not 0 = note on
				{
					cobQueue.put(eventInfo->time, eventInfo->data[0]  & 0x0F, eventInfo->data[1], eventInfo->data[2]);
				}
				else
				{
					// velocity = 0 = note mute = note off)
					CobChunkStruct* cobInfo;
					while ((cobInfo = cobQueue.find(eventInfo->data[0]  & 0x0F, eventInfo->data[1])) != 0)
					{
						cobInfo->noteDuration = eventInfo->time - cobInfo->eventTime;
					}
				}
				break;

			case 0xA0:
				break;

			case 0xB0:
				cobQueue.put(eventInfo->time, eventInfo->data[0], eventInfo->data[1], eventInfo->data[2]);
				break;

			case 0xC0:
				cobQueue.put(eventInfo->time, eventInfo->data[0], eventInfo->data[1], 0xFF);
				break;

			case 0xD0:
				cobQueue.put(eventInfo->time, eventInfo->data[0], eventInfo->data[1], 0xFF);
				break;

			case 0xE0:
				cobQueue.put(eventInfo->time, eventInfo->data[0], eventInfo->data[1], eventInfo->data[2]);
				break;

			default:
				printf("Unknown command 0x%02X, %04X\n", command, trackOffsets[i]); 
			}
		}
	}

	// Process the last piece of lyric
	if (prevLyricEventInfo != 0)
	{
		temp = 2;
		int textLength = ReadVarLen(prevLyricEventInfo->data, &temp);

		for (int i = 0; i < textLength; i++)
		{
			lyricQueue.put(prevLyricEventInfo->time, prevLyricEventInfo->data[temp + i]);
		}
	}

	// Scale the time
	float scale = (float)(((float)tempo / header.Division) / 17000.0);

	// Create the lyric header
	char lheader[]  = "H0002\r@@00@13\r@";
	char ltitle[]   = "5150\r";
	char lheader2[] = "@\r@@";
	char lauthor[]  = "Edward Van Halen\r";
	int uncompressLyricSize = 0;
	unsigned char* lyricPart = new unsigned char[lyricQueue.size * 4 + 256];

	bool lyricFirstTime = true;
	int convertChar;
	for (i = 0; i < (int)strlen(lheader); i++)
	{
		lyricPart[uncompressLyricSize++] = lheader[i];
		if (lheader[i] == 0x0D)
			lyricPart[uncompressLyricSize++] = 0x0A;
		else
			lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;
	}
	for (i = 0; i < (int)strlen(ltitle); i++)
	{
		convertChar = convertASCIIToJBKFont(ltitle[i], lyricFirstTime);
		if (convertChar != 0)
		{
			lyricPart[uncompressLyricSize++] = convertChar;
			if (convertChar == 0x0D)
				lyricPart[uncompressLyricSize++] = 0x0A;
			else
				lyricPart[uncompressLyricSize++] = 0;
			lyricPart[uncompressLyricSize++] = 0;
			lyricPart[uncompressLyricSize++] = 0;
		}

		lyricFirstTime = false;
	}
	convertChar = convertASCIIToJBKFont(0, lyricFirstTime);
	if (convertChar != 0)
	{
		lyricPart[uncompressLyricSize++] = convertChar;
		if (convertChar == 0x0D)
			lyricPart[uncompressLyricSize++] = 0x0A;
		else
			lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;

		lyricFirstTime = false;
	}
	for (i = 0; i < (int)strlen(lheader2); i++)
	{
		lyricPart[uncompressLyricSize++] = lheader2[i];
		if (lheader2[i] == 0x0D)
			lyricPart[uncompressLyricSize++] = 0x0A;
		else
			lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;
	}
	lyricFirstTime = true;
	for (i = 0; i < (int)strlen(lauthor); i++)
	{
		convertChar = convertASCIIToJBKFont(lauthor[i], lyricFirstTime);
		if (convertChar)
		{
			lyricPart[uncompressLyricSize++] = convertChar;
			if (convertChar == 0x0D)
				lyricPart[uncompressLyricSize++] = 0x0A;
			else
				lyricPart[uncompressLyricSize++] = 0;
			lyricPart[uncompressLyricSize++] = 0;
			lyricPart[uncompressLyricSize++] = 0;
		}

		lyricFirstTime = false;
	}
	convertChar = convertASCIIToJBKFont(0, lyricFirstTime);
	if (convertChar != 0)
	{
		lyricPart[uncompressLyricSize++] = convertChar;
		if (convertChar == 0x0D)
			lyricPart[uncompressLyricSize++] = 0x0A;
		else
			lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;
		lyricPart[uncompressLyricSize++] = 0;
	}

	// Create lyric buffer
	CobLyricChunkStruct* lyricInfo;
	int scaledTime;
	lyricFirstTime = true;
	while ((lyricInfo = lyricQueue.get()) != 0)
	{
		convertChar = convertASCIIToJBKFont(lyricInfo->syl, lyricFirstTime);
		if (convertChar != 0)
		{
			lyricPart[uncompressLyricSize++] = convertChar;
			if (convertChar == 0x0D)
				lyricPart[uncompressLyricSize++] = 0x0A;
			else
				lyricPart[uncompressLyricSize++] = 0;

			if (0)//convertChar == ' ' || convertChar == 0x0D)
			{
				lyricPart[uncompressLyricSize++] = 0;// 0 time is okay, but two 0xD with zero time indicate end of lyric
				lyricPart[uncompressLyricSize++] = 0;
			}
			else
			{
				scaledTime = (int)(lyricInfo->eventTime * scale);
				lyricPart[uncompressLyricSize++] = scaledTime & 0xFF;
				lyricPart[uncompressLyricSize++] = scaledTime >> 8;
			}
		}

		lyricFirstTime = false;
	}

	convertChar = convertASCIIToJBKFont(0, lyricFirstTime);
	if (convertChar != 0)
	{
		lyricPart[uncompressLyricSize++] = convertChar;
		if (convertChar == 0x0D)
			lyricPart[uncompressLyricSize++] = 0x0A;
		else
			lyricPart[uncompressLyricSize++] = 0;

			if (0)//convertChar == ' ' || convertChar == 0x0D)
			{
				lyricPart[uncompressLyricSize++] = 0;
				lyricPart[uncompressLyricSize++] = 0;
			}
			else
			{
				lyricPart[uncompressLyricSize++] = scaledTime & 0xFF;
				lyricPart[uncompressLyricSize++] = scaledTime >> 8;
			}
	}

	// Add two charet return at the end to signal end of lyric
	lyricPart[uncompressLyricSize++] = 0x0D;
	lyricPart[uncompressLyricSize++] = 0x0A;
	lyricPart[uncompressLyricSize++] = 0x00;
	lyricPart[uncompressLyricSize++] = 0x00;

	lyricPart[uncompressLyricSize++] = 0x0D;
	lyricPart[uncompressLyricSize++] = 0x0A;
	lyricPart[uncompressLyricSize++] = 0x00;
	lyricPart[uncompressLyricSize++] = 0x00;

	// Create MIDI buffer
	unsigned char* midiPart = new unsigned char[cobQueue.size * 5];
	CobChunkStruct* cobInfo;
	int prevTimeLine = 0;
	int uncompressMidiSize = 0;
	while ((cobInfo = cobQueue.get()) != 0)
	{
		int scaledTime = (int)(cobInfo->eventTime * scale);

		if (cobInfo->eventTime <= prevTimeLine)
			midiPart[uncompressMidiSize++] = 0;
		else
			midiPart[uncompressMidiSize++] = scaledTime - prevTimeLine;

		if (cobInfo->command > 0x0F) // 16 channels max 
			midiPart[uncompressMidiSize++] = 1;
		else
			midiPart[uncompressMidiSize++] = (unsigned char)(cobInfo->noteDuration * scale);

		midiPart[uncompressMidiSize++] = cobInfo->command;
		midiPart[uncompressMidiSize++] = cobInfo->note;
		midiPart[uncompressMidiSize++] = cobInfo->velocity;

		prevTimeLine = scaledTime;
	}

	// Compress lyric part
	unsigned char* lyricPartCompress = new unsigned char[uncompressLyricSize];
	int compressLyricSize = compress(lyricPart, uncompressLyricSize, lyricPartCompress);
/*
	// Test code
	FILE* fpTest = fopen("00001_lyric.com", "wb");
	fwrite(lyricPartCompress, 1, compressLyricSize, fpTest);
	fclose(fpTest);
*/
/*
	// Test code
	FILE* fpTest = fopen("00001_midi.unc", "wb");
	fwrite(midiPart, 1, uncompressMidiSize, fpTest);
	fclose(fpTest);
*/
	// Compress the MIDI part
	unsigned char* midiPartCompress = new unsigned char[uncompressMidiSize];
	int compressMidiSize = compress(midiPart, uncompressMidiSize, midiPartCompress);
/*
	// Test code
	FILE* fpTest = fopen("00001_midi.com", "wb");
	fwrite(midiPartCompress, 1, compressMidiSize, fpTest);
	fclose(fpTest);
*/
	// Write COB file

	// Compress lyric size
	temp = compressLyricSize;
	correctEndian(&temp, 1);
	fileSize += fwrite(&temp, 1, sizeof(temp), outFile);
	
	// Uncompress lyric size
	temp = uncompressLyricSize;
	correctEndian(&temp, 1);
	fileSize += fwrite(&temp, 1, sizeof(temp), outFile);
	
	// Unknown number
	temp = 0x00;
	correctEndian(&temp, 1);
	fileSize += fwrite(&temp, 1, sizeof(temp), outFile);

	fileSize += fwrite(lyricPartCompress, 1, compressLyricSize, outFile);
	
	// Compress MIDI size
	temp = compressMidiSize;
	correctEndian(&temp, 1);
	fileSize += fwrite(&temp, 1, sizeof(temp), outFile);

	// Uncompress MIDI size
	temp = uncompressMidiSize;
	correctEndian(&temp, 1);
	fileSize += fwrite(&temp, 1, sizeof(temp), outFile);

	// Unknown number
	temp = 0x00;
	correctEndian(&temp, 1);
	fileSize += fwrite(&temp, 1, sizeof(temp), outFile);

	fileSize += fwrite(midiPartCompress, 1, compressMidiSize, outFile);

	// Pad to 16 divisible
	unsigned char padding[16];
	memset(padding, 0, sizeof(padding));
	int totalSize = compressLyricSize + compressMidiSize + 24;
	if ((totalSize % 16) != 0)
	{
		fileSize += fwrite(padding, 1, 16 - (totalSize % 16), outFile);
	}

	delete[] lyricPart;
	delete[] lyricPartCompress;
	
	delete[] midiPart;
	delete[] midiPartCompress;

	// Clean up
	for (i = 0; i < header.NumTracks; i++)
	{
		delete[] tracks[i];
	}

	delete[] tracks;
	delete[] trackOffsets;

	for (i = 0; i < 256; i++)
		delete[] asciiJBKLookup[i];

	delete[] asciiJBKLookup;

	return fileSize;
}
